use clap::Parser;
use image::{GenericImageView, ImageFormat, Pixel, RgbaImage};
use num_traits::AsPrimitive;
use serde::{Deserialize, Serialize};
use std::path::PathBuf;

#[inline(always)]
fn tile_size() -> u32 {
	32
}

/// Take a spritesheet and split into individual sprites, skipping empty space
#[derive(Parser, Serialize, Deserialize, Clone, Debug)]
#[clap(author, version = "0.8.0")]
pub struct Split {
	/// The path to the spritesheet file
	#[serde(default)]
	pub input: String,
	/// The base path for all of the created sprites. Should be a directory, which will be created if it does not exist
	#[serde(default)]
	pub output: String,

	/// The amount of horizontal space between each sprite in the image
	#[clap(short = 'x', long, default_value_t = 0)]
	#[serde(default)]
	space_x: u32,
	/// The amount of vertical space between each sprite in the image
	#[clap(short = 'y', long, default_value_t = 0)]
	#[serde(default)]
	space_y: u32,
	/// The amount of space around the edge of the image
	#[clap(short, long, default_value_t = 0)]
	#[serde(default)]
	padding: u32,
	/// The size of each tile in the spritesheet. Assumed to be square tiles
	#[clap(short, long)]
	#[serde(default)]
	tile_size: Option<u32>,
	#[clap(short = 'w', long)]
	#[serde(default)]
	tile_width: Option<u32>,
	#[clap(short = 'h', long)]
	#[serde(default)]
	tile_height: Option<u32>,
	/// When set, files will be numbered consecutively, regardless of empty sprites. By default, files
	/// will be numbered based on their inferred "tile id", based on position in the sheet
	#[clap(short, long, default_value_t = false)]
	#[serde(default)]
	sequential: bool,
}

fn params_to_columns(
	dimension: impl AsPrimitive<f32>,
	tile_size: impl AsPrimitive<f32>,
	padding: impl AsPrimitive<f32>,
	spacing: impl AsPrimitive<f32>,
) -> u32 {
	let dimension = dimension.as_();
	let tile_size = tile_size.as_();
	let padding = padding.as_();
	let spacing = spacing.as_();

	let base = (dimension - (2.0 * padding) - spacing) / (tile_size + spacing);
	base.floor() as u32 + if spacing == 0.0 { 0 } else { 1 }
}

impl Split {
	pub fn run(&self, image: &RgbaImage) -> anyhow::Result<()> {
		log::info!(
			"Image loaded with size {} x {}",
			image.width(),
			image.height()
		);

		let tile_width = self.tile_width.or(self.tile_size).ok_or_else(|| {
			anyhow::Error::msg(
				"Must provide either '--tile-size' or '--tile-x' flags to specify tile width",
			)
		})?;
		let tile_height = self.tile_height.or(self.tile_size).ok_or_else(|| {
			anyhow::Error::msg(
				"Must provide either '--tile-size' or '--tile-y' flags to specify tile height",
			)
		})?;

		let columns = params_to_columns(image.width(), tile_width, self.padding, self.space_x);
		let rows = params_to_columns(image.height(), tile_height, self.padding, self.space_y);
		log::info!("Inferred sheet contains {} columns", columns);
		log::info!("Inferred sheet contains {} rows", rows);

		std::fs::create_dir_all(&self.output)?;
		let path_base = PathBuf::from(&self.output);

		let mut file_count = 0;
		for x in 0..columns {
			for y in 0..rows {
				let img_x = self.padding + (x * tile_width) + (x * self.space_x);
				let img_y = self.padding + (y * tile_height) + (y * self.space_y);
				// let img_y = y * self.tile_size;

				let view = image.view(img_x, img_y, tile_width, tile_height);

				if view
					.pixels()
					.any(|(_, _, pixel)| pixel.channels().get(3).map(|c| c > &0).unwrap_or(false))
				{
					view.to_image().save_with_format(
						path_base.join(format!("{}.png", file_count)),
						ImageFormat::Png,
					)?;

					if self.sequential {
						file_count += 1;
					}
				}

				if !self.sequential {
					file_count += 1;
				}
			}
		}

		Ok(())
	}
}

#[cfg(test)]
mod tests {
	#[test]
	fn no_pad_space() {
		let columns = super::params_to_columns(160, 16, 0, 0);
		let rows = super::params_to_columns(160, 16, 0, 0);

		assert_eq!(columns, 10);
		assert_eq!(rows, 10);

		let columns = super::params_to_columns(512, 16, 0, 0);
		let rows = super::params_to_columns(160, 16, 0, 0);

		assert_eq!(columns, 32);
		assert_eq!(rows, 10);
	}

	#[test]
	fn some_pad() {
		let columns = super::params_to_columns(168, 16, 4, 0);
		let rows = super::params_to_columns(168, 16, 4, 0);

		assert_eq!(columns, 10);
		assert_eq!(rows, 10);

		let columns = super::params_to_columns(520, 16, 4, 0);
		let rows = super::params_to_columns(168, 16, 4, 0);

		assert_eq!(columns, 32);
		assert_eq!(rows, 10);
	}
	#[test]
	fn some_space() {
		let columns = super::params_to_columns(168, 16, 0, 0);
		let rows = super::params_to_columns(168, 16, 0, 0);

		assert_eq!(columns, 10);
		assert_eq!(rows, 10);

		let columns = super::params_to_columns(520, 16, 0, 0);
		let rows = super::params_to_columns(168, 16, 0, 0);

		assert_eq!(columns, 32);
		assert_eq!(rows, 10);
	}
	#[test]
	fn different_spacing() {
		let space_x = 3;
		let space_y = 2;

		let columns = super::params_to_columns(189, 13, 0, space_x);
		let rows = super::params_to_columns(133, 13, 0, space_y);

		assert_eq!(columns, 12);
		assert_eq!(rows, 9);
	}

	#[test]
	fn correct_conversion() {
		let columns = super::params_to_columns(434, 32, 3, 1);
		let rows = super::params_to_columns(302, 32, 6, 1);

		assert_eq!(columns, 13);
		assert_eq!(rows, 9);
	}
}
